import { existsSync, readFileSync, writeFileSync } from "fs";
import path from "node:path";
import slug from "slug";
import type Vault from "./vault.js";
import {
  combineMarkdownAndFrontmatter,
  extractFrontmatter,
  extractMarkdown,
  extractTags,
  type FrontmatterShape,
} from "./markdown.js";

export default class MarkdownDocument {
  /** A reference to the vault containing this document */
  readonly vault: Vault;

  /** The original file path to this document */
  readonly path: string;

  /** The directory/folder that this Document is in, relative to the vault root */
  readonly dirname: string;

  /** An array of directory names this document is found under */
  readonly taxonomy: string[];

  /** The name of this document's file, without the file extension */
  readonly filename: string;

  /** A url-safe version of this file's name */
  readonly slug: string;

  /** The raw content of the file */
  private _content: string;

  /** The history of this files contents after each modification */
  readonly contentHistory: string[] = [];

  /** The parsed frontmatter of the file */
  private _frontmatter: FrontmatterShape;

  /** Structured document metadata */
  meta: Record<string, unknown> = {};

  /** The markdown portion of the file (without frontmatter) */
  private _markdown: string;

  tags: string[];

  constructor(filePath: string, vault: Vault) {
    if (!existsSync(filePath)) {
      throw new Error(`File not found: ${filePath}`);
    }

    const rawFileContent = readFileSync(filePath, "utf-8");
    const fileDirname = path.dirname(filePath);
    const vaultPath = vault.vaultPath;
    const frontmatter = extractFrontmatter(rawFileContent);

    this.path = filePath;
    this.dirname = path.relative(vaultPath, fileDirname);
    this.taxonomy = this.dirname.split(path.sep);
    this.filename = path.basename(filePath, ".md");
    this.slug = slug(`${this.taxonomy.join("-")}-${this.filename}`);
    this._content = rawFileContent;
    this.contentHistory = [rawFileContent];
    this._frontmatter = frontmatter;
    this._markdown = extractMarkdown(rawFileContent);
    this.vault = vault;

    // Identify tags from content and frontmatter
    const frontmatterTags = frontmatter.tags;
    const contentTags = extractTags(this._markdown);
    this.tags = frontmatterTags.concat(contentTags);
  }

  toJSON() {
    return {
      path: this.path,
      dirname: this.dirname,
      taxonomy: this.taxonomy,
      filename: this.filename,
      slug: this.slug,
      frontmatter: this._frontmatter,
      markdown: this._markdown,
    };
  }

  set markdown(newValue: string) {
    this._markdown = newValue;
    this._content = combineMarkdownAndFrontmatter(newValue, this.frontmatter);
    this.contentHistory.push(this._content);
  }

  /** The markdown portion of the file (without frontmatter) */
  get markdown() {
    return this._markdown;
  }

  set frontmatter(newValue: FrontmatterShape) {
    this._frontmatter = newValue;
    this._content = combineMarkdownAndFrontmatter(
      this._markdown,
      this._frontmatter,
    );
    this.contentHistory.push(this._content);
  }

  /** The parsed frontmatter of the file */
  get frontmatter() {
    return this._frontmatter;
  }

  /** The raw content of the file */
  get content() {
    return this._content;
  }

  set content(newValue: string) {
    this._content = newValue;
    this._frontmatter = extractFrontmatter(newValue);
    this._markdown = extractMarkdown(newValue);
    this.contentHistory.push(this._content);
  }

  hasTag(tag: string) {
    return this.tags.includes(tag);
  }

  hasTaxonomy(dirs: string[]) {
    if (dirs.length < this.taxonomy.length) {
      return false;
    }

    for (let i = 0; i < dirs.length; i++) {
      if (dirs[i] !== this.taxonomy[i]) {
        return false;
      }
    }

    return true;
  }

  write() {
    writeFileSync(
      this.path,
      combineMarkdownAndFrontmatter(this._markdown, this.frontmatter),
      "utf-8",
    );
  }
}
